@@ -90,6 +90,10 @@ typedef double ev_tstamp;
 /*****************************************************************************/
 
 #if EV_STAT_ENABLE
+# ifdef _WIN32
+#  include <time.h>
+#  include <sys/types.h>
+# endif
 # include <sys/stat.h>
 #endif
 
@@ -423,7 +427,6 @@ void ev_set_syserr_cb (void (*cb)(const char *msg));
 extern struct ev_loop *ev_default_loop_ptr;
 extern struct ev_loop *ev_default_loop_init (unsigned int flags);
 
-
 EV_INLINE struct ev_loop *
 ev_default_loop_uc (void)
 {
