@@ -1,6 +1,6 @@
 /*****
 *
-* Copyright (C) 2004, 2005 PreludeIDS Technologies. All Rights Reserved.
+* Copyright (C) 2004-2007,2008 PreludeIDS Technologies. All Rights Reserved.
 * Author: Yoann Vandoorselaere <yoann.v@prelude-ids.com>
 *
 * This file is part of the Prelude-Manager program.
@@ -37,7 +37,6 @@
 #include <libprelude/prelude-connection-pool.h>
 
 #include "reverse-relaying.h"
-#include "server-logic.h"
 #include "server-generic.h"
 #include "sensor-server.h"
 #include "manager-options.h"
@@ -45,11 +44,6 @@
 #include "sensor-server.h"
 
 
-typedef struct {
-        pthread_mutex_t mutex;
-        prelude_connection_pool_t *pool;
-} reverse_relay_t;
-
 
 struct reverse_relay_receiver {
         prelude_list_t list;
@@ -58,20 +52,31 @@ struct reverse_relay_receiver {
         uint64_t analyzerid;
 
         unsigned int count;
-        pthread_mutex_t mutex;
+
         prelude_failover_t *failover;
         server_generic_client_t *client;
 };
 
 
-static PRELUDE_LIST(receiver_list);
+typedef struct {
+        prelude_list_t list;
+
+        uint64_t analyzerid;
+        prelude_msg_t *msg;
+} mqueue_t;
+
+
 static prelude_msgbuf_t *msgbuf;
-static pthread_mutex_t receiver_mutex = PTHREAD_MUTEX_INITIALIZER;
-static reverse_relay_t initiator = { PTHREAD_MUTEX_INITIALIZER, NULL};
+static PRELUDE_LIST(mqueue_list);
+static pthread_mutex_t mqueue_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+
+static PRELUDE_LIST(receiver_list);
+static pthread_mutex_t receiver_list_mutex = PTHREAD_MUTEX_INITIALIZER;
 
 extern manager_config_t config;
 extern prelude_client_t *manager_client;
-
+static prelude_connection_pool_t *initiator = NULL;
 
 
 static int connection_event_cb(prelude_connection_pool_t *pool,
@@ -103,15 +108,16 @@ static reverse_relay_receiver_t *get_next_receiver(prelude_list_t **iter)
         prelude_list_t *tmp;
         reverse_relay_receiver_t *rrr = NULL;
 
-        pthread_mutex_lock(&receiver_mutex);
-
+        /*
+         * Locking here is not required since the list is never
+         * modified by the worker thread. We only protect writing
+         * of the list, and reading it from the worker thread.
+         */
         prelude_list_for_each_continue_safe(&receiver_list, tmp, *iter) {
                 rrr = prelude_list_entry(tmp, reverse_relay_receiver_t, list);
                 break;
         }
 
-        pthread_mutex_unlock(&receiver_mutex);
-
         return rrr;
 }
 
@@ -125,14 +131,10 @@ int reverse_relay_set_receiver_alive(reverse_relay_receiver_t *rrr, server_gener
         prelude_msg_t *msg;
         prelude_failover_t *failover = rrr->failover;
 
-        pthread_mutex_lock(&rrr->mutex);
-
         size = prelude_failover_get_saved_msg(failover, &msg);
         if ( size == 0 )
                 rrr->client = client;
 
-        pthread_mutex_unlock(&rrr->mutex);
-
         if ( size < 0 ) {
                 prelude_perror((prelude_error_t) size, "could not retrieve saved message from disk");
                 return -1;
@@ -153,7 +155,7 @@ int reverse_relay_set_receiver_alive(reverse_relay_receiver_t *rrr, server_gener
                         return ret;
                 }
 
-                server_logic_notify_write_enable((server_logic_client_t *) client);
+                server_generic_notify_write_enable(client);
                 return 1;
         }
 
@@ -174,10 +176,8 @@ int reverse_relay_set_initiator_dead(prelude_connection_t *cnx)
 {
         int ret = -1;
 
-        pthread_mutex_lock(&initiator.mutex);
-        if ( initiator.pool )
-                ret = prelude_connection_pool_set_connection_dead(initiator.pool, cnx);
-        pthread_mutex_unlock(&initiator.mutex);
+        if ( initiator )
+                ret = prelude_connection_pool_set_connection_dead(initiator, cnx);
 
         return ret;
 }
@@ -186,9 +186,7 @@ int reverse_relay_set_initiator_dead(prelude_connection_t *cnx)
 
 void reverse_relay_set_receiver_dead(reverse_relay_receiver_t *rrr)
 {
-        pthread_mutex_lock(&rrr->mutex);
         rrr->client = NULL;
-        pthread_mutex_unlock(&rrr->mutex);
 }
 
 
@@ -196,7 +194,7 @@ void reverse_relay_set_receiver_dead(reverse_relay_receiver_t *rrr)
 int reverse_relay_new_receiver(reverse_relay_receiver_t **rrr, server_generic_client_t *client, uint64_t analyzerid)
 {
         int ret;
-        char buf[512];
+        char fname[PATH_MAX];
         reverse_relay_receiver_t *new;
 
         new = malloc(sizeof(*new));
@@ -207,22 +205,19 @@ int reverse_relay_new_receiver(reverse_relay_receiver_t **rrr, server_generic_cl
         new->client = client;
         new->analyzerid = analyzerid;
 
-        prelude_client_profile_get_backup_dirname(prelude_client_get_profile(manager_client), buf, sizeof(buf));
-        snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), "/%" PRELUDE_PRIu64, analyzerid);
+        prelude_client_profile_get_backup_dirname(prelude_client_get_profile(manager_client), fname, sizeof(fname));
+        snprintf(fname + strlen(fname), sizeof(fname) - strlen(fname), "/%" PRELUDE_PRIu64, analyzerid);
 
-        ret = prelude_failover_new(&new->failover, buf);
+        ret = prelude_failover_new(&new->failover, fname);
         if ( ret < 0 ) {
                 prelude_perror(ret, "could not create failover");
                 free(new);
                 return -1;
         }
 
-        pthread_mutex_init(&new->mutex, NULL);
-
-        pthread_mutex_lock(&receiver_mutex);
+        pthread_mutex_lock(&receiver_list_mutex);
         prelude_list_add_tail(&receiver_list, &new->list);
-        pthread_mutex_unlock(&receiver_mutex);
-
+        pthread_mutex_unlock(&receiver_list_mutex);
         *rrr = new;
 
         return 0;
@@ -247,29 +242,20 @@ reverse_relay_receiver_t *reverse_relay_search_receiver(uint64_t analyzerid)
 
 static int send_msgbuf(prelude_msgbuf_t *msgbuf, prelude_msg_t *msg)
 {
-        int ret;
-        reverse_relay_receiver_t *item = prelude_msgbuf_get_data(msgbuf);
+        mqueue_t *mq;
 
-        pthread_mutex_lock(&item->mutex);
-
-        if ( item->client ) {
-                /*
-                 * We cannot safely write to the gnutls session that is
-                 * shared by another thread, thus queue the message for
-                 * processing.
-                 */
-                sensor_server_queue_write_client(item->client, msg);
-                pthread_mutex_unlock(&item->mutex);
-
-                return 0;
+        mq = malloc(sizeof(*mq));
+        if ( ! mq ) {
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted: %s.\n", strerror(errno));
+                return -1;
         }
 
-        pthread_mutex_unlock(&item->mutex);
-        ret = prelude_failover_save_msg(item->failover, msg);
-        if ( ret < 0 )
-                prelude_perror(ret, "could not save message to disk");
+        mq->msg = msg;
+        mq->analyzerid = *(uint64_t *) prelude_msgbuf_get_data(msgbuf);
 
-        prelude_msg_destroy(msg);
+        pthread_mutex_lock(&mqueue_mutex);
+        prelude_list_add_tail(&mqueue_list, &mq->list);
+        pthread_mutex_unlock(&mqueue_mutex);
 
         return 0;
 }
@@ -319,46 +305,106 @@ static int get_issuer_analyzerid(idmef_message_t *idmef, uint64_t *analyzerid)
 
 
 
+static mqueue_t *mqueue_get_next(void)
+{
+        mqueue_t *q = NULL;
+
+        pthread_mutex_lock(&mqueue_mutex);
+
+        if ( prelude_list_is_empty(&mqueue_list) )
+                goto out;
+
+        q = prelude_list_entry(mqueue_list.next, mqueue_t, list);
+        prelude_list_del(&q->list);
+
+out:
+        pthread_mutex_unlock(&mqueue_mutex);
+        return q;
+}
+
+
+
 void reverse_relay_send_receiver(idmef_message_t *idmef)
 {
         int ret;
         uint64_t analyzerid;
-        prelude_list_t *iter = NULL;
-        reverse_relay_receiver_t *item;
+        prelude_bool_t empty;
 
+        /*
+         * If there is no receiver, no need to queue the message.
+         */
+        pthread_mutex_lock(&receiver_list_mutex);
+        empty = prelude_list_is_empty(&receiver_list);
+        pthread_mutex_unlock(&receiver_list_mutex);
+
+        if ( empty )
+                return;
+
+        /*
+         * Create a new item in the message queue, containing
+         * the message to be sent, as well as the analyzerid of the
+         * emitter.
+         */
         ret = get_issuer_analyzerid(idmef, &analyzerid);
         if ( ret < 0 )
                 return;
 
         /*
-         * FIXME: this is dirty since we are encoding
-         * the IDMEF message once per receiver.
-         *
-         * Implement a better scheme where we create an intermediate
-         * "linking" object, which share a refcount, a message, and
-         * a mutex, but hold it's list member private.
+         * Convert from idmef_message_t to prelude_msg_t,
+         * this will trigger the msgbuf callback where an mqueue_t
+         * object will be created, and attached to the list of message
+         * to be emited.
          */
-        while ( (item = get_next_receiver(&iter)) ) {
+        prelude_msgbuf_set_data(msgbuf, &analyzerid);
+        idmef_message_write(idmef, msgbuf);
+        prelude_msgbuf_mark_end(msgbuf);
 
-                if ( analyzerid == item->analyzerid )
-                        /* we're not an echo server, let's skip the sender */
-                        continue;
-
-                prelude_msgbuf_set_data(msgbuf, item);
-                idmef_message_write(idmef, msgbuf);
-                prelude_msgbuf_mark_end(msgbuf);
-        }
+        /*
+         * Finally, restart the main server event loop so that it
+         * take into account the event to be written, and call
+         * reverse_relay_send_prepared().
+         */
+        server_generic_notify_event();
 }
 
 
 
+void reverse_relay_send_prepared(void)
+{
+        int ret;
+        mqueue_t *mq;
+        prelude_list_t *iter = NULL;
+        reverse_relay_receiver_t *receiver;
+
+        while ( (mq = mqueue_get_next()) ) {
+
+                while ( (receiver = get_next_receiver(&iter)) ) {
+
+                        if ( mq->analyzerid == receiver->analyzerid )
+                                continue;
+
+                        if ( receiver->client )
+                                sensor_server_write_client(receiver->client, prelude_msg_ref(mq->msg));
+                        else {
+                                ret = prelude_failover_save_msg(receiver->failover, mq->msg);
+                                if ( ret < 0 )
+                                        prelude_perror(ret, "could not save message to disk");
+                        }
+                }
+
+                prelude_msg_destroy(mq->msg);
+                free(mq);
+        }
+}
+
+
 static void destroy_current_initiator(void)
 {
         sensor_fd_t *client;
         prelude_list_t *tmp;
         prelude_connection_t *cnx;
 
-        prelude_list_for_each(prelude_connection_pool_get_connection_list(initiator.pool), tmp) {
+        prelude_list_for_each(prelude_connection_pool_get_connection_list(initiator), tmp) {
                 cnx = prelude_linked_object_get_object(tmp);
 
                 client = prelude_connection_get_data(cnx);
@@ -369,8 +415,8 @@ static void destroy_current_initiator(void)
                 }
         }
 
-        prelude_connection_pool_destroy(initiator.pool);
-        initiator.pool = NULL;
+        prelude_connection_pool_destroy(initiator);
+        initiator = NULL;
 }
 
 
@@ -382,33 +428,30 @@ int reverse_relay_create_initiator(const char *arg)
 
         cp = prelude_client_get_profile(manager_client);
 
-        pthread_mutex_lock(&initiator.mutex);
-        if ( initiator.pool )
+        if ( initiator )
                 destroy_current_initiator();
 
-        ret = prelude_connection_pool_new(&initiator.pool, cp, PRELUDE_CONNECTION_PERMISSION_IDMEF_READ);
+        ret = prelude_connection_pool_new(&initiator, cp, PRELUDE_CONNECTION_PERMISSION_IDMEF_READ);
         if ( ret < 0 )
                 goto out;
 
-        prelude_connection_pool_set_flags(initiator.pool, PRELUDE_CONNECTION_POOL_FLAGS_RECONNECT);
-        prelude_connection_pool_set_event_handler(initiator.pool, PRELUDE_CONNECTION_POOL_EVENT_DEAD |
+        prelude_connection_pool_set_flags(initiator, PRELUDE_CONNECTION_POOL_FLAGS_RECONNECT);
+        prelude_connection_pool_set_event_handler(initiator, PRELUDE_CONNECTION_POOL_EVENT_DEAD |
                                                   PRELUDE_CONNECTION_POOL_EVENT_ALIVE, connection_event_cb);
 
-        ret = prelude_connection_pool_set_connection_string(initiator.pool, arg);
+        ret = prelude_connection_pool_set_connection_string(initiator, arg);
         if ( ret < 0 ) {
-                prelude_connection_pool_destroy(initiator.pool);
+                prelude_connection_pool_destroy(initiator);
                 goto out;
         }
 
-        ret = prelude_connection_pool_init(initiator.pool);
+        ret = prelude_connection_pool_init(initiator);
         if ( ret < 0 ) {
-                prelude_connection_pool_destroy(initiator.pool);
+                prelude_connection_pool_destroy(initiator);
                 goto out;
         }
 
  out:
-        pthread_mutex_unlock(&initiator.mutex);
-
         return ret;
 }
 
